/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/* Chaotic pendulum                                                          */
/* ================                                                          */
/*                                                                           */
/* A twodimensional frictionless chaotic pendulum consisting of two          */
/* connected rotating homogenous bar masses with different lengths           */
/* and initial angles.                                                       */
/*                                                                           */
/* The author doesn't know if the equations are equavilent also for          */
/* for massless bars with point masses at the end of the bars. If mass       */
/* circles are displayed they are shown at the outer ends of the bars.       */
/*                                                                           */
/* Set graphics resolution, console mode or window mode, background colour   */  
/* etc and the pendulum parameters in the file "chaopend.cfg".               */
/*                                                                           */
/*                                                                           */
/* Author etc                                                                */
/* ==========                                                                */
/*                                                                           */
/* Program version: chaopend.c 1.00                                          */
/*                                                                           */
/* The program is free to copy and modify.                                   */
/*                                                                           */
/* Author: Stefan Spaennare, Lund, Sweden                                    */
/*                                                                           */
/* E-mail: stefan@spaennare.se                                               */
/*                                                                           */
/* Home page:    http://www.spaennare.se/index.html                          */
/* Program page: http://www.spaennare.se/ssprog2.html                        */
/*                                                                           */
/* C version for Windows: 2006-07-04;   Using Allegro game library and       */
/*                                      Microsoft Visual C++ 6.0.            */
/* Latest update:         2006-07-13                                         */
/*                                                                           */
/*                                                                           */
/* References                                                                */
/* ==========                                                                */
/*                                                                           */
/* 1. The chaotic pendulum model and the equations for it were found at      */
/*    the University museum at Kulturen in Lund, Sweden.                     */
/*                                                                           */
/* 2. The program uses a classical 4th order Runge-Kutta method for          */
/*    integration of the differential equations.                             */
/*                                                                           */
/* 3. The program uses a random number generator ran2() from Numerical       */
/*    Recipes in C Second Edition.                                           */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "allegro.h"

double pi=3.14159265358979323;

double timetic=(double)(CLOCKS_PER_SEC);

char cfg_name[21]="chaopend.cfg";

char image_name[21]="chaopend.bmp";


#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(idum)
long *idum;
{
   int j;
   long k;
   static long idum2=123456789;
   static long iy=0; 
   static long iv[NTAB];
   double temp;

   if (*idum <= 0) {
      if (-(*idum) < 1) {
         *idum=1;
      }
      else {
         *idum=-(*idum);
      } /* if */
      idum2=(*idum);
      for (j=NTAB+7; j>=0; j--) {
         k=(*idum)/IQ1;
	 *idum=IA1*(*idum-k*IQ1)-k*IR1;
	 if (*idum < 0) {
	    *idum += IM1;
	 } /* if */
	 if (j < NTAB) {
	    iv[j]=*idum;
	 } /* if */
      } /* for j */
      iy=iv[0];
   } /* if */
   k=(*idum)/IQ1;
   *idum=IA1*(*idum-k*IQ1)-k*IR1;
   if (*idum < 0) {
      *idum += IM1;
   } /* if */
   k=idum2/IQ2;
   idum2=IA2*(idum2-k*IQ2)-k*IR2;
   if (idum2 < 0) {
      idum2 += IM2;
   } /* if */
   j=iy/NDIV;
   iy=iv[j]-idum2;
   iv[j] = *idum;
   if (iy < 1) {
      iy += IMM1;
   } /* if */
   temp=AM*(double)(iy);
   if (temp > RNMX) {
      return(RNMX);
   }
   else {
      return(temp);
   } /* if */

} /* ran2 */


void comp_acc(acc1,acc2)
double *acc1;
double *acc2;
{
   double v1,v2,m1,m2,l1,l2,g,a1,a2,a3;
   double d1,d1t,d2,d2t,sind1d2,deninv;
     
      d1=acc1[1];  d2=acc1[2];
      d1t=acc1[3]; d2t=acc1[4];
      l1=acc1[5];  l2=acc1[6];
      m1=acc1[7];  m2=acc1[8];
      g=acc1[9];

      sind1d2=sin(d1-d2);

      v1=-(m1+m2)*g*l1*sin(d1)-m2*l1*l2*d2t*d2t*sind1d2;
      v2=-m2*g*l2*sin(d2)+m2*l1*l2*d1t*d1t*sind1d2;

      a1=(m1+m2)*l1*l1;
      a2=m2*l1*l2*cos(d1-d2);
      a3=m2*l2*l2;

      deninv=1.0/(a1*a3-a2*a2);

      acc2[1]=deninv*(a3*v1-a2*v2);
      acc2[2]=deninv*(a1*v2-a2*v1);

} /* comp_acc */


int main(argc,argv)
int argc;
char *argv[];
{

   FILE *cfgfile;

   BITMAP *bmp;
   PALETTE pal;

   int i,j,seed,delay0;
   int xsize,ysize,col,c0;
   int wintype,imsave;
   int x0i,y0i,x1i,y1i,x2i,y2i;
   int x1i2,y1i2,x2i2,y2i2;
   int x1s,y1s,x2s,y2s;
   int col11,col12,col21,col22,bgcol,bordercol,standcol;
   int dispflag11,dispflag12,dispflag21,dispflag22;
   int borderflag,rcmi1,rcmi2;
   int dispstep,standflag,flag;

   long idum;

   double xrf,yrf,xs0,ys0,sum,totsum,ttt1,ttt2;
   double m1,m2,l1,l2,g,d10,d20,lfac,lf;
   double x1,y1,x2,y2,w,rcm1,rcm2;
   double d1,d1t,d2,d2t,dt;
   double h,h2,aa1,bb1,cc1,dd1,k11,k21,inv3;
   double aa2,bb2,cc2,dd2,k12,k22;
   double beta1,delta1,beta2,delta2;
   double sfac,standw,standh,temp;

   double acc1[11];
   double acc2[11];

   int cr[21];
   int cg[21];
   int cb[21];

   time_t tt;

   clock_t time1,timediff1;


   if ((cfgfile = fopen(cfg_name,"r")) == NULL) {
      printf("File %s not found.\n",cfg_name);
      exit(0);
   } /* if */


   fscanf(cfgfile,"%d %d",&xsize,&ysize);
   fscanf(cfgfile,"%d",&wintype);
   fscanf(cfgfile,"%d",&imsave);

   fscanf(cfgfile,"%d",&bgcol);
   fscanf(cfgfile,"%d",&bordercol);
   fscanf(cfgfile,"%d",&standcol);
   fscanf(cfgfile,"%d",&col11);
   fscanf(cfgfile,"%d",&col12);
   fscanf(cfgfile,"%d",&col21);
   fscanf(cfgfile,"%d",&col22);

   fscanf(cfgfile,"%d",&borderflag);
   fscanf(cfgfile,"%d",&standflag);
   fscanf(cfgfile,"%d",&dispflag11);
   fscanf(cfgfile,"%d",&dispflag12);
   fscanf(cfgfile,"%d",&dispflag21);
   fscanf(cfgfile,"%d",&dispflag22);
   fscanf(cfgfile,"%d",&dispstep);

   fscanf(cfgfile,"%lf",&sfac);
   fscanf(cfgfile,"%lf",&standw);
   fscanf(cfgfile,"%lf",&standh);

   fscanf(cfgfile,"%d",&delay0);
   fscanf(cfgfile,"%lf",&dt);
   fscanf(cfgfile,"%lf",&g);

   fscanf(cfgfile,"%lf",&d10);
   fscanf(cfgfile,"%lf",&d20);

   fscanf(cfgfile,"%lf",&lfac);
   fscanf(cfgfile,"%lf",&l1);
   fscanf(cfgfile,"%lf",&l2);
   fscanf(cfgfile,"%lf",&rcm1);
   fscanf(cfgfile,"%lf",&rcm2);

   fscanf(cfgfile,"%lf",&m1);
   fscanf(cfgfile,"%lf",&m2);


   if ((wintype < 0) || (wintype > 1)) {
      printf("wintype must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((imsave < 0) || (imsave > 1)) {
      printf("imsave must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((borderflag < 0) || (borderflag > 1)) {
      printf("borderflag must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((standflag < 0) || (standflag > 1)) {
      printf("standflag must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((dispflag11 < 0) || (dispflag11 > 3)) {
      printf("dispflag11 must be 0, 1, 2 or 3.\n");
      exit(0);
   } /* if */

   if ((dispflag12 < 0) || (dispflag12 > 1)) {
      printf("dispflag12 must be 0 or 1.\n");
      exit(0);
   } /* if */

   if ((dispflag21 < 0) || (dispflag21 > 3)) {
      printf("dispflag21 must be 0, 1, 2 or 3.\n");
      exit(0);
   } /* if */

   if ((dispflag22 < 0) || (dispflag22 > 1)) {
      printf("dispflag22 must be 0 or 1.\n");
      exit(0);
   } /* if */


   allegro_init();

   install_keyboard();

   set_color_depth(32);


   if (wintype == 0) {
      if (set_gfx_mode(GFX_AUTODETECT,xsize,ysize,0,0) < 0) {
         if (set_gfx_mode(GFX_AUTODETECT_WINDOWED,xsize,ysize,0,0) < 0) {
            allegro_exit();
            printf("Graphics error or wrong resolution.\n");
            printf("Try to set resolution to 640 480.\n");
            exit(0);
         } /* if */
      } /* if */
   } /* if */

   if (wintype == 1) {
      if (set_gfx_mode(GFX_AUTODETECT_WINDOWED,xsize,ysize,0,0) < 0) {
         if (set_gfx_mode(GFX_AUTODETECT,xsize,ysize,0,0) < 0) {
            allegro_exit();
            printf("Graphics error or wrong resolution.\n");
            printf("Try to set resolution to 640 480.\n");
            exit(0);
         } /* if */
      } /* if */
   } /* if */


   set_palette(desktop_palette);


   cr[0]=0;    cg[0]=0;    cb[0]=0;
   cr[1]=255;  cg[1]=255;  cb[1]=255;

   cr[2]=255;  cg[2]=0;    cb[2]=0;
   cr[3]=0;    cg[3]=255;  cb[3]=0;
   cr[4]=0;    cg[4]=0;    cb[4]=255;

   cr[5]=255;  cg[5]=0;    cb[5]=255;
   cr[6]=255;  cg[6]=255;  cb[6]=0;
   cr[7]=0;    cg[7]=255;  cb[7]=255;

   cr[8]=128;  cg[8]=0;    cb[8]=0;
   cr[9]=0;    cg[9]=128;  cb[9]=0;
   cr[10]=0;   cg[10]=0;   cb[10]=128;

   cr[11]=128; cg[11]=128; cb[11]=0;
   cr[12]=128; cg[12]=0;   cb[12]=128;
   cr[13]=0;   cg[13]=128; cb[13]=128;

   cr[14]=128; cg[14]=128; cb[14]=128;
   cr[15]=160; cg[15]=160; cb[15]=160;
   cr[16]=192; cg[16]=192; cb[16]=192;


   c0=bgcol;
   col=makecol(cr[c0],cg[c0],cb[c0]);

   for (i=0; i<ysize; i++) {
      line(screen,0,i,xsize-1,i,col);
   } /* for i */ 


   tt=time(0);
   ttt1=(double)(tt)+0.5;
   ttt2=fmod(ttt1,2147483647.0);
   seed=(long)(ttt2);
   idum=-seed-1;


   xrf=(double)(xsize)/640.0;
   yrf=(double)(ysize)/480.0;
   xs0=320.0;
   ys0=240.0;

   x0i=(int)(xrf*xs0+0.5);
   y0i=(int)(2.0*(1.0-sfac)*yrf*ys0+0.5);
   x1i=x0i;
   y1i=y0i;
   x2i=x0i;
   y2i=y0i;

   x1s=(int)(2.0*standw*sfac*xrf*xs0+0.5);
   y1s=(int)(2.0*(1.0-sfac)*yrf*ys0-2.0*standh*sfac*yrf*ys0+0.5);
   x2s=-x1s+x0i;
   x1s=x1s+x0i;
   y2s=y1s;

   lf=2.0*lfac*sfac*ys0/(l1+l2);

   rcmi1=(int)(lf*yrf*rcm1+0.5);
   rcmi2=(int)(lf*yrf*rcm2+0.5);


   if (d10 < 0.0) {
      temp=fabs(d10);
      d10=(180.0-temp)*ran2(&idum)+temp;
   } /* if */

   if (d20 < 0.0) {
      temp=fabs(d20);
      d20=(180.0-temp)*ran2(&idum)+temp;
   } /* if */

   d1=d10*pi/180.0;
   d2=d20*pi/180.0;
   d1t=0.0;
   d2t=0.0;

   acc1[5]=l1; acc1[6]=l2;
   acc1[7]=m1; acc1[8]=m2;
   acc1[9]=g;

   h=dt;
   h2=0.5*dt;
   
   inv3=1.0/3.0;

   w=-90.0*pi/180.0;

   totsum=0.0;

   flag=0;
    
   i=1;


   do {

      
      /* Start of 4th order Runge-Kutta integration. */


      acc1[1]=d1;  acc1[2]=d2;
      acc1[3]=d1t; acc1[4]=d2t;
      
      comp_acc(acc1,acc2);

      aa1=h2*acc2[1];
      aa2=h2*acc2[2];

      beta1=h2*(d1t+0.5*aa1);
      beta2=h2*(d2t+0.5*aa2);

      acc1[1]=d1+beta1; acc1[2]=d2+beta2;
      acc1[3]=d1t+aa1;  acc1[4]=d2t+aa2;

      comp_acc(acc1,acc2);

      bb1=h2*acc2[1];
      bb2=h2*acc2[2];

      acc1[1]=d1+beta1; acc1[2]=d2+beta2;
      acc1[3]=d1t+bb1;  acc1[4]=d2t+bb2;
      
      comp_acc(acc1,acc2);

      cc1=h2*acc2[1];
      cc2=h2*acc2[2];

      delta1=h*(d1t+cc1);
      delta2=h*(d2t+cc2);

      acc1[1]=d1+delta1;   acc1[2]=d2+delta2;
      acc1[3]=d1t+2.0*cc1; acc1[4]=d2t+2.0*cc2;

      comp_acc(acc1,acc2);

      dd1=h2*acc2[1];
      dd2=h2*acc2[2];

      k11=inv3*(aa1+bb1+cc1);
      k12=inv3*(aa2+bb2+cc2);

      k21=inv3*(aa1+2.0*bb1+2.0*cc1+dd1);
      k22=inv3*(aa2+2.0*bb2+2.0*cc2+dd2);

      d1=d1+h*(d1t+k11);
      d2=d2+h*(d2t+k12);

      d1t=d1t+k21;
      d2t=d2t+k22;


      /* End of 4th order Runge-Kutta integration. */


      /* Start of display modes. */


      if (i == dispstep) {


         x1=l1*cos(d1+w);
         y1=l1*sin(d1+w);

         x2=l2*cos(d2+w);
         y2=l2*sin(d2+w); 

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag12 == 1) {
            circle(screen,x1i,y1i,rcmi1,col);
         } /* if */

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag22 == 1) {
            circle(screen,x2i,y2i,rcmi2,col);
         } /* if */

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag11 == 1) {
            line(screen,x0i,y0i,x1i,y1i,col);
         } /* if */

         c0=bgcol;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag21 == 1) {
            line(screen,x1i,y1i,x2i,y2i,col);
         } /* if */

         x1i=(int)(xrf*lf*x1+0.5)+x0i;
         y1i=-(int)(yrf*lf*y1+0.5)+y0i;

         x2i=(int)(xrf*lf*x2+0.5)+x1i;
         y2i=-(int)(yrf*lf*y2+0.5)+y1i;

         if (standflag == 1) {
            c0=standcol;
            col=makecol(cr[c0],cg[c0],cb[c0]);
            line(screen,x1s,ysize-1,x1s,y1s,col);
            line(screen,x2s,ysize-1,x2s,y2s,col);
            line(screen,x1s,y1s,x2s,y2s,col);
         } /* if */

         if (flag == 0) {
            x1i2=x1i;
            y1i2=y1i;
            x2i2=x2i;
            y2i2=y2i;
         } /* if */

         c0=col11;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag11 == 1) {
            line(screen,x0i,y0i,x1i,y1i,col);
         } /* if */

         if (dispflag11 == 2) {
            putpixel(screen,x1i,y1i,col);
         } /* if */

         if (dispflag11 == 3) {
            line(screen,x1i2,y1i2,x1i,y1i,col);
         } /* if */

         c0=col21;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag21 == 1) {
            line(screen,x1i,y1i,x2i,y2i,col);
         } /* if */

         if (dispflag21 == 2) {
            putpixel(screen,x2i,y2i,col);
         } /* if */

         if (dispflag21 == 3) {
            line(screen,x2i2,y2i2,x2i,y2i,col);
         } /* if */

         c0=col12;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag12 == 1) {
            circle(screen,x1i,y1i,rcmi1,col);
         } /* if */

         c0=col22;
         col=makecol(cr[c0],cg[c0],cb[c0]);

         if (dispflag22 == 1) {
            circle(screen,x2i,y2i,rcmi2,col);
         } /* if */

         if (borderflag == 1) {
            c0=bordercol;
            col=makecol(cr[c0],cg[c0],cb[c0]);
            line(screen,0,0,xsize-1,0,col);
            line(screen,xsize-1,0,xsize-1,ysize-1,col);
            line(screen,xsize-1,ysize-1,0,ysize-1,col);
            line(screen,0,ysize-1,0,0,col);
         } /* if */

         x1i2=x1i;
         y1i2=y1i;
         x2i2=x2i;
         y2i2=y2i;

         flag=1;

	 i=0;

      } /* if */


      /* End of display modes. */

      i++;

      sum=0.0;

      for (j=1; j<=delay0; j++) {
         sum=sum+1.0/(double)(j);
      } /* for j */

      totsum=totsum+sum;

   } while (!keypressed());



   if (imsave == 1) {
      get_palette(pal);
      bmp=create_sub_bitmap(screen,0,0,xsize,ysize);
      save_bitmap(image_name,bmp,pal);
      destroy_bitmap(bmp);
   } /* if */

   allegro_exit();

   fclose(cfgfile);

   printf("dummy = %12.6f\n",totsum);

   return(0);

} /* End  */

END_OF_MAIN();
